# Number Theory utils
import random
import secrets
from math import gcd


def miller_rabin(n, runs=40):
    # Implementation uses the Miller-Rabin Primality Test
    # The optimal number of rounds for this test is 40
    # See http://stackoverflow.com/questions/6325576/how-many-iterations-of-rabin-miller-should-i-use-for-cryptographic-safe-primes
    # for justification

    if n == 2:
        return True

    # If number is even, it's a composite number
    if not n & 1:
        return False

    r, s = 0, n - 1
    while s % 2 == 0:
        r += 1
        s //= 2
    for _ in range(runs):
        a = random.randrange(3, n - 1, 2)
        x = pow(a, s, n)
        if x == 1 or x == n - 1:
            continue
        for _ in range(r - 1):
            x = pow(x, 2, n)
            if x == n - 1:
                break
        else:
            return False
    return True


def random_prime(bits):
    while True:
        n = secrets.randbits(bits)
        if miller_rabin(n):
            return n


def random_coprime(n):
    assert n > 1
    while True:
        e = random.randrange(1, n)
        if gcd(n, e) == 1:
            return e


def xgcd(b, n):
    """Takes positive integers a, b as input, and return a triple (g, x, y), such that ax + by = g = gcd(a, b)"""
    x0, x1, y0, y1 = 1, 0, 0, 1
    while n != 0:
        q, b, n = b // n, n, b % n
        x0, x1 = x1, x0 - q * x1
        y0, y1 = y1, y0 - q * y1
    return b, x0, y0


def mulinv(b, n):
    """An application of extended GCD algorithm to finding modular inverses"""
    g, x, _ = xgcd(b, n)
    assert g == 1, 'Numbers must be coprimes'
    return x % n


def legendre(a, p):
    """https://en.wikipedia.org/wiki/Legendre_symbol"""
    assert miller_rabin(p), f"{p} is not a prime"
    mod = pow(a, (p - 1) // 2, p)
    return -1 if mod == p - 1 else mod


def modsqrt(a, p):
    """
        https://eli.thegreenplace.net/2009/03/07/computing-modular-square-roots-in-python
        Find a quadratic residue (mod p) of 'a'. p
        must be an odd prime.
        Solve the congruence of the form:
            x^2 = a (mod p)
        And returns x. Note that p - x is also a root.
        0 is returned is no square root exists for
        these a and p.
        The Tonelli-Shanks algorithm is used (except
        for some simple cases in which the solution
        is known from an identity). This algorithm
        runs in polynomial time (unless the
        generalized Riemann hypothesis is false).
    """
    # Simple cases
    #
    if legendre(a, p) != 1:
        return 0
    elif a == 0:
        return 0
    elif p == 2:
        return p
    elif p % 4 == 3:
        return pow(a, (p + 1) // 4, p)

    # Partition p-1 to s * 2^e for an odd s (i.e.
    # reduce all the powers of 2 from p-1)
    #
    s = p - 1
    e = 0
    while s % 2 == 0:
        s //= 2
        e += 1

    # Find some 'n' with a legendre symbol n|p = -1.
    # Shouldn't take long.
    #
    n = 2
    while legendre(n, p) != -1:
        n += 1

    # Here be dragons!
    # Read the paper "Square roots from 1; 24, 51,
    # 10 to Dan Shanks" by Ezra Brown for more
    # information
    #

    # x is a guess of the square root that gets better
    # with each iteration.
    # b is the "fudge factor" - by how much we're off
    # with the guess. The invariant x^2 = ab (mod p)
    # is maintained throughout the loop.
    # g is used for successive powers of n to update
    # both a and b
    # r is the exponent - decreases with each update
    #
    x = pow(a, (s + 1) // 2, p)
    b = pow(a, s, p)
    g = pow(n, s, p)
    r = e

    while True:
        t = b
        m = 0
        for m in range(r):
            if t == 1:
                break
            t = pow(t, 2, p)

        if m == 0:
            return x

        gs = pow(g, 2 ** (r - m - 1), p)
        g = (gs * gs) % p
        x = (x * gs) % p
        b = (b * g) % p
        r = m



from enum import Enum, unique

from hdtools.opcodes import AddressType


@unique
class NETWORK(Enum):
    BTC_MAIN = 'btc'
    BTC_TEST = 'btct'


btc_main = {
    'hrp': 'bc',
    'keyhash': b'\x00',
    'scripthash': b'\x05',
    'wif': b'\x80',
    'extended_prv': {
        # https://github.com/spesmilo/electrum-docs/blob/master/xpub_version_bytes.rst
        AddressType.P2PKH: b'\x04\x88\xad\xe4',  # xprv
        AddressType.P2WPKH: b'\x04\xb2\x43\x0c',  # zprv
        AddressType.P2WSH: b'\x02\xaa\x7a\x99',  # Zprv
        AddressType.P2WPKH_P2SH: b'\x04\x9d\x78\x78',  # yprv
        AddressType.P2WSH_P2SH: b'\x02\x95\xb4\x3f'  # Yprv
    },
    'extended_pub': {
        # https://github.com/spesmilo/electrum-docs/blob/master/xpub_version_bytes.rst
        AddressType.P2PKH: b'\x04\x88\xb2\x1e',  # xpub
        AddressType.P2WPKH: b'\x04\xb2\x47\x46',  # zpub
        AddressType.P2WSH: b'\x02\xaa\x7e\xd3',  # Zpub
        AddressType.P2WPKH_P2SH: b'\x04\x9d\x7c\xb2',  # ypub
        AddressType.P2WSH_P2SH: b'\x02\x95\xb4\x3f'  # Ypub
    },
    'utxo_url': 'https://blockchain.info/unspent?active={address}',
    'rawtx_url': 'https://blockchain.info/rawtx/{txid}?format=hex',
    'broadcast_url': 'https://blockchain.info/pushtx'

}

btc_test = {
    'hrp': 'tb',
    'keyhash': b'\x6f',
    'scripthash': b'\xc4',
    'wif': b'\xef',
    'extended_prv': {
        # https://github.com/spesmilo/electrum-docs/blob/master/xpub_version_bytes.rst
        AddressType.P2PKH: b'\x04\x35\x83\x94',  # tprv
        AddressType.P2WPKH: b'\x04\x5f\x18\xbc',  # vprv
        AddressType.P2WSH: b'\x02\x57\x50\x48',  # Vprv
        AddressType.P2WPKH_P2SH: b'\x04\x4a\x4e\x28',  # uprv
        AddressType.P2WSH_P2SH: b'\x02\x42\x85\xb5'  # Uprv
    },
    'extended_pub': {
        # https://github.com/spesmilo/electrum-docs/blob/master/xpub_version_bytes.rst
        AddressType.P2PKH: b'\x04\x35\x87\xcf',  # tpub
        AddressType.P2WPKH: b'\x04\x5f\x1c\xf6',  # vpub
        AddressType.P2WSH: b'\x02\x57\x54\x83',  # Vpub
        AddressType.P2WPKH_P2SH: b'\x04\x4a\x52\x62',  # upub
        AddressType.P2WSH_P2SH: b'\x02\x42\x89\xef'  # Upub
    },
    'utxo_url': 'https://testnet.blockchain.info/unspent?active={address}',
    'rawtx_url': 'https://testnet.blockchain.info/rawtx/{txid}?format=hex',
    'broadcast_url': 'https://testnet.blockchain.info/pushtx'
}

networks = {
    NETWORK.BTC_MAIN: btc_main,
    NETWORK.BTC_TEST: btc_test
}


def get_network_attr(attr, network='btc'):
    return networks[NETWORK(network)][attr]



from enum import Enum


class AddressType(Enum):
    P2PK = 'P2PK'
    P2PKH = 'P2PKH'
    P2SH = 'P2SH'
    P2WPKH = 'P2WPKH'
    P2WSH = 'P2WSH'
    P2WPKH_P2SH = 'P2WPKH-P2SH'
    P2WSH_P2SH = 'P2WSH-P2SH'



from hdtools.conversions import int_to_bytes, hex_to_bytes, bytes_to_int, bytes_to_hex

import base64
from hashlib import sha256 as hash_function


class Message:
    """Basic data class with useful constructors and methods"""

    def __init__(self, bts):
        self.msg = bts

    @classmethod
    def from_int(cls, i):
        return cls(int_to_bytes(i))

    @classmethod
    def from_hex(cls, h):
        return cls(hex_to_bytes(h))

    @classmethod
    def from_str(cls, s, encoding='utf-8'):
        return cls(str.encode(s, encoding))

    @classmethod
    def from_binary(cls, b):
        return cls(int_to_bytes(int(b, 2)))

    @classmethod
    def from_base64(cls, s):
        return cls(base64.b64decode(s))

    @classmethod
    def from_file(cls, path):
        with open(path, 'rb') as f:
            b = f.read()
        return cls(b)

    def int(self):
        return bytes_to_int(self.msg)

    def str(self, encoding='utf-8'):
        return self.msg.decode(encoding)

    def hex(self):
        return bytes_to_hex(self.msg)

    def bin(self):
        return format(bytes_to_int(self.msg), 'b')

    def bytes(self):
        return self.msg

    def base64(self):
        return base64.b64encode(self.msg).decode()

    def __repr__(self):
        return repr(self.msg)

    def __eq__(self, other):
        return self.msg == other.msg

    def __len__(self):
        return len(self.bin())  # in bits

    def hash(self):
        return hash_function(self.msg).hexdigest()



def bytes_to_int(bts):
    return int.from_bytes(bts, 'big')


def int_to_bytes(i):
    length = max(1, (i.bit_length() + 7) // 8)
    return i.to_bytes(length, 'big')


def hex_to_int(h):
    return int(h, 16)


def str_to_bytes(s):
    return str.encode(s, 'utf-8')


def str_to_int(s):
    return bytes_to_int(str_to_bytes(s))


def int_to_str(i):
    return bytes_to_str(int_to_bytes(i))


def bytes_to_str(b):
    return b.decode('utf-8')


def int_to_hex(i):
    return format(i, 'x')


def bytes_to_hex(b):
    return b.hex()


def hex_to_bytes(h):
    return bytes.fromhex(h)


def str_to_hex(s):
    return bytes_to_hex(str_to_bytes(s))


def hex_to_str(h):
    return bytes_to_str(hex_to_bytes(h))



import setuptools
from distutils.core import setup

setup(
    name='hdtools',
    version='0.1.1',
    description='hd-tools for bitcoin and bitcoin-test',
    author='Mahdi Fooladgar',
    author_email='fooladgar@morvarid.io',
    url='https://github.com/morvaridio/hdtools',
    license='MIT',
    packages=[
        'hdtools',
    ]
)



from ecdsa import SigningKey, SECP256k1 as DefaultCurve
from ecdsa.ellipticcurve import Point

from base58 import b58decode, b58encode

from hdtools.conversions import hex_to_bytes, bytes_to_hex, int_to_bytes, bytes_to_int, hex_to_int
from hdtools.message import Message as BaseMessage
from hdtools.network import get_network_attr

from hdtools.nt_utils import modsqrt
from hdtools.crypto_utils import sha256


def f(x, curve=DefaultCurve.curve):
    """
    Compute y**2 = x^3 + ax + b in field FP
    :param x: x
    :param curve: The curve
    :return: result of relation
    """
    return (x ** 3 + curve.a() * x + curve.b()) % curve.p()


def ecdsa_point_creator(x, y):
    return Point(
        curve=DefaultCurve.curve,
        x=x,
        y=y
    )


class PrivateKey(BaseMessage):
    def __init__(self, bts, network='btc'):
        super().__init__(bts)
        self.network = network
        self._key = SigningKey.from_string(
            bts,
            curve=DefaultCurve
        )

    @staticmethod
    def random(network='btc'):
        return PrivateKey(
            SigningKey.generate(curve=DefaultCurve).to_string(),
            network=network
        )

    @staticmethod
    def from_wif(wif: str, network='btc') -> "PrivateKey":
        bts = b58decode(wif)
        network_byte, key, checksum = bts[0:1], bts[1:-4], bts[-4:]

        assert sha256(sha256(network_byte + key))[:4] == checksum, 'Invalid Checksum'
        assert network_byte == get_network_attr('wif', network), 'Invalid Network byte'

        if key.endswith(b'\x01'):
            key = key[:-1]
            compressed = True  # TODO
        else:
            compressed = False  # TODO
        return PrivateKey(key)

    def wif(self, compressed=False):
        extended = get_network_attr('wif', self.network) + self.bytes() + (b'\x01' if compressed else b'')
        hashed = sha256(sha256(extended))
        checksum = hashed[:4]
        return b58encode(extended + checksum)

    def to_public(self):
        point = self._key.get_verifying_key().pubkey.point
        return PublicKey(point, self.network)

    def __repr__(self):
        return f"PrivateKey({self.msg})"

    def sign_hash(self, digest):
        return self._key.sign_digest(digest)


class PublicKey:
    def __init__(self, point, network):
        self.network = network
        self.point = point

    def __eq__(self, other):
        return self.point == other.point

    def __repr__(self):
        return f"PublicKey({self.x()}, {self.y()})"

    def x(self):
        return self.point.x()

    def y(self):
        return self.point.y()

    @staticmethod
    def from_private(private, network='btc'):
        private_key = PrivateKey.from_int(private) if isinstance(private, int) else private
        private_key.network = network
        return private_key.to_public()

    @staticmethod
    def decode(key: bytes, network='btc'):  # TODO Easier implementation
        if key.startswith(b'\x04'):  # uncompressed key
            assert len(key) == 65, 'An uncompressed public key must be 65 bytes long'
            x, y = bytes_to_int(key[1:33]), bytes_to_int(key[33:])
        else:  # compressed key
            assert len(key) == 33, 'A compressed public key must be 33 bytes long'
            x = bytes_to_int(key[1:])
            root = modsqrt(f(x), DefaultCurve.curve.p())
            if key.startswith(b'\x03'):  # odd root
                y = root if root % 2 == 1 else -root % DefaultCurve.curve.p()
            elif key.startswith(b'\x02'):  # even root
                y = root if root % 2 == 0 else -root % DefaultCurve.curve.p()
            else:
                assert False, 'Wrong key format'

        return PublicKey(ecdsa_point_creator(x, y), network=network)

    @staticmethod
    def from_hex(hex_string: str, network='btc'):
        return PublicKey.decode(hex_to_bytes(hex_string), network)

    def encode(self, compressed=False) -> bytes:  # TODO Maybe easier Implementation
        if compressed:
            if self.y() & 1:  # odd root
                return b'\x03' + int_to_bytes(self.x()).rjust(32, b'\x00')
            else:  # even root
                return b'\x02' + int_to_bytes(self.x()).rjust(32, b'\x00')
        return b'\x04' + int_to_bytes(self.x()).rjust(32, b'\x00') + int_to_bytes(self.y()).rjust(32, b'\x00')

    def hex(self, compressed=False) -> str:
        return bytes_to_hex(self.encode(compressed=compressed))

    def to_address(self, address_type, compressed=None):
        from hdtools.address import Address
        if compressed is not False and address_type == 'P2PKH':
            return Address.from_public_key(self, address_type, compressed=True)
        return Address.from_public_key(self, address_type)



from base58 import b58encode

from hdtools import bech32
from hdtools.keys import PublicKey
from hdtools.crypto_utils import hash160
from hdtools.network import get_network_attr
from hdtools.script import witness_byte, push
from hdtools.crypto_utils import sha256
from hdtools.opcodes import AddressType


def legacy_address(public_key: PublicKey, version_byte: bytes) -> str:
    """https://en.bitcoin.it/wiki/Technical_background_of_version_1_Bitcoin_addresses"""
    bts = public_key.encode(compressed=False) if isinstance(public_key, PublicKey) else public_key
    hashed = hash160(bts)
    payload = version_byte + hashed
    return hashed_payload_to_address(payload)


def hashed_payload_to_address(payload):
    checksum = sha256(sha256(payload))[:4]
    address = payload + checksum
    return b58encode(address)


def pubkey_to_bech32(public_key: PublicKey, witver: int) -> str:
    """https://github.com/bitcoin/bips/blob/master/bip-0141.mediawiki#witness-program"""
    witprog = hash160(public_key.encode(compressed=True))
    return bech32.encode(
        get_network_attr('hrp', public_key.network),
        witver,
        witprog)


class Address:
    @staticmethod
    def from_p2pkh(public_key: PublicKey, compressed=False) -> 'str':
        return legacy_address(
            public_key.encode(compressed=True) if compressed else public_key,
            version_byte=get_network_attr('keyhash', public_key.network))

    @staticmethod
    def from_p2wpkh_p2sh(public_key: PublicKey) -> 'str':
        return legacy_address(
            witness_byte(witver=0) + push(hash160(public_key.encode(compressed=True))),
            version_byte=get_network_attr('scripthash', public_key.network)
        )

    @staticmethod
    def from_p2wpkh(public_key: PublicKey) -> 'str':
        return pubkey_to_bech32(public_key, witver=0x00)

    @staticmethod
    def from_public_key(public_key: PublicKey, version='P2PKH', compressed=False) -> 'str':
        key_to_addr_versions = {
            AddressType.P2PKH: Address.from_p2pkh,
            AddressType.P2WPKH_P2SH: Address.from_p2wpkh_p2sh,
            AddressType.P2WPKH: Address.from_p2wpkh
        }

        if version == AddressType.P2PKH.value:
            return key_to_addr_versions[AddressType(version)](public_key, compressed)

        return key_to_addr_versions[AddressType(version)](public_key)



"""
Implementation of BIP32, BIP49, BIP84, ...
References:
    https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki#child-key-derivation-functions
    https://iancoleman.io/bip39/
"""
import hashlib
from typing import Union

from base58 import b58encode, b58decode
import hmac

from mnemonic import Mnemonic

from hdtools.conversions import bytes_to_int, int_to_bytes, bytes_to_hex, hex_to_bytes
from hdtools.network import get_network_attr
from hdtools.opcodes import AddressType
from hdtools.keys import PrivateKey, PublicKey, DefaultCurve
from hdtools.crypto_utils import hash160, sha256, sha512

Key = Union[PrivateKey, PublicKey]


class ExtendedKey:
    root_path = NotImplemented

    def __init__(self, key: Key, code: bytes, depth=0, i=None, parent=b'\x00\x00\x00\x00', path=None,
                 address_type='P2PKH'):
        self.key = key
        self.code = code

        assert depth in range(256), 'Depth can only be 0-255'
        self.depth = depth
        if i is not None:
            assert 0 <= i < 1 << 32, f'Invalid i: {i}'
        self.i = i
        self.parent = parent
        self.path = path or self.root_path

        assert (
                       self.depth == 0 and
                       self.i is None and
                       self.parent == b'\x00\x00\x00\x00' and
                       self.path == self.root_path
               ) or (
                       self.depth != 0 and
                       self.i is not None and
                       self.parent != b'\x00\x00\x00\x00' and
                       self.path != self.root_path
               ), f"Unable to determine if root path (" \
            f"depth={self.depth}, i={self.i}, " \
            f"path={self.path}, " \
            f"parent={bytes_to_hex(self.parent)})"

        self.type = AddressType(address_type)

    def child(self, i):
        raise NotImplementedError

    def is_master(self):
        return self.depth == 0 and \
               self.i is None and \
               self.parent == b'\x00\x00\x00\x00' and \
               self.path == self.root_path

    def __truediv__(self, other):
        if isinstance(other, float):
            # hardened child derivation
            i = int(other) + 2 ** 31
        elif isinstance(other, int):
            # non-hardened child derivation
            i = other
        else:
            raise TypeError
        return self.child(i)

    def __floordiv__(self, other):
        if not isinstance(other, int):
            raise TypeError
        return self.child(other + 2 ** 31)

    def id(self):
        raise NotImplementedError

    def fingerprint(self):
        return self.id()[:4]

    def serialize(self):
        raise NotImplementedError

    def encode(self):
        data = self.serialize()
        assert len(data) == 78
        checksum = sha256(sha256(data))[:4]
        return b58encode(data + checksum)

    @classmethod
    def deserialize(cls, bts: bytes, network='btc'):
        def read(n):
            nonlocal bts
            data, bts = bts[:n], bts[n:]
            return data

        net = read(4)
        is_private = net in get_network_attr('extended_prv', network).values()
        is_public = net in get_network_attr('extended_pub', network).values()
        assert is_public ^ is_private, f'Invalid network bytes : {bytes_to_hex(net)}'
        address_lookup = {val: key for key, val in (
            get_network_attr('extended_prv', network) if is_private else get_network_attr('extended_pub',
                                                                                          network)).items()}
        constructor = XPrv if is_private else XPub
        depth = bytes_to_int(read(1))
        assert depth in range(256), f'Invalid depth : {depth}'
        fingerprint = read(4)
        i = bytes_to_int(read(4))
        if depth == 0:
            i = None
            path = None
        else:
            ih = f'{i}' if i < 2 ** 31 else f"{i - 2 ** 31}h"
            path = '/'.join([constructor.root_path] + ['x' for _ in range(depth - 1)] + [ih])

        code = read(32)
        key = read(33)
        key = PrivateKey(key, network=network) if is_private else PublicKey.decode(key, network=network)
        assert not bts, 'Leftover bytes'
        return constructor(key, code, depth=depth, i=i, parent=fingerprint, path=path, addresstype=address_lookup[net])

    @classmethod
    def decode(cls, string: str, network='btc'):
        bts = b58decode(string)
        assert len(bts) == 82, f'Invalid length {len(bts)}'
        data, checksum = bts[:78], bts[78:]
        assert sha256(sha256(data)).startswith(checksum), 'Invalid checksum'
        return cls.deserialize(data, network)

    def __eq__(self, other):
        return self.encode() == other.encode


class KeyDerivationError(Exception):  # TODO
    pass


class XPrv(ExtendedKey):
    root_path = 'm'

    def child(self, i) -> 'XPrv':
        hardened = i >= 1 << 31

        if hardened:
            I = hmac.new(
                key=self.code,
                msg=self.key_data() + int_to_bytes(i).rjust(4, b'\x00'),
                digestmod=hashlib.sha512
            ).digest()
        else:
            I = hmac.new(
                key=self.code,
                msg=self.key.to_public().encode(compressed=True) + int_to_bytes(i).rjust(4, b'\x00'),
                digestmod=hashlib.sha512
            ).digest()

        I_L, I_R = bytes_to_int(I[:32]), I[32:]
        key = (I_L + self.key.int()) % DefaultCurve.order

        if I_L >= DefaultCurve.order or key == 0:
            return self.child(i + 1)

        ret_code = I_R
        if hardened:
            path = self.path + f'/{i - 2 ** 31}h'
        else:
            path = self.path + f'/{i}'

        private = PrivateKey.from_int(key)
        private.network = self.key.network
        return XPrv(
            key=private,
            code=ret_code,
            depth=self.depth + 1,
            i=i,
            parent=self.fingerprint(),
            path=path,
            address_type=self.type.value
        )

    def to_xpub(self) -> 'XPub':
        return XPub(
            self.key.to_public(),
            self.code,
            depth=self.depth,
            i=self.i,
            parent=self.parent,
            path=self.path.replace('m', 'M'),
            address_type=self.type.value
        )

    def to_child_xpub(self, i: int) -> 'XPub':
        # return self.child(i).to_xpub()  # works always
        return self.to_xpub().child(i)  # works only for non-hardened child keys

    def id(self):
        return hash160(self.key.to_public().encode(compressed=True))

    def key_data(self):
        return self.key.bytes().rjust(33, b'\x00')

    def serialize(self):
        version = get_network_attr('extended_prv', self.key.network)[self.type]
        depth = int_to_bytes(self.depth)
        child = bytes(4) if self.is_master() else int_to_bytes(self.i).rjust(4, b'\x00')
        return version + depth + self.parent + child + self.code + self.key_data()

    @staticmethod
    def from_seed(seed: Union[bytes, str], address_type='P2PKH', network='btc') -> 'XPrv':
        if isinstance(seed, str):
            seed = hex_to_bytes(seed)
        assert 16 <= len(seed) <= 64, 'Seed should be between 128 and 512 bits'

        I = hmac.new(key=b"Bitcoin seed", msg=seed, digestmod=hashlib.sha512).digest()
        I_L, I_R = I[:32], I[32:]
        if bytes_to_int(I_L) == 0 or bytes_to_int(I_L) > DefaultCurve.order:
            raise KeyDerivationError

        key, code = PrivateKey(I_L, network=network), I_R
        return XPrv(key, code, address_type=address_type)

    @staticmethod
    def from_mnemonic(mnemonic: str, pass_phrase='', address_type='P2PKH', network='btc'):
        seed = Mnemonic.to_seed(mnemonic, pass_phrase)
        return XPrv.from_seed(seed, address_type, network)

    def address(self, address_type=None):
        return self.key.to_public().to_address(address_type or self.type.value, compressed=True)


class XPub(ExtendedKey):
    root_path = 'M'

    def child(self, i: int) -> 'XPub':
        hardened = i >= 1 << 31

        if hardened:
            raise KeyDerivationError('Cannot derive a hardened key from an extended public key')

        I = hmac.new(key=self.code, msg=self.key_data() + int_to_bytes(i).rjust(4, b'\x00'),
                     digestmod=hashlib.sha512).digest()

        I_L, I_R = I[:32], I[32:]

        key = PrivateKey(I_L).to_public().point + self.key.point
        ret_code = I_R
        path = self.path + f'/{i}'

        # TODO add point at infinity check
        return XPub(
            PublicKey(key, network=self.key.network),
            ret_code,
            depth=self.depth + 1,
            i=i,
            parent=self.fingerprint(),
            path=path,
            address_type=self.type.value
        )

    def id(self):
        return hash160(self.key.encode(compressed=True))

    def key_data(self):
        return self.key.encode(compressed=True)

    def serialize(self):
        version = get_network_attr('extended_pub', self.key.network)[self.type]
        depth = int_to_bytes(self.depth)
        child = bytes(4) if self.is_master() else int_to_bytes(self.i).rjust(4, b'\x00')
        return version + depth + self.parent + child + self.code + self.key_data()

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self.path}, key={self.key.hex(compressed=True)}"

    def address(self, address_type=None):
        return self.key.to_address(address_type or self.type.value, compressed=True)
